/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */



/*****************************************************************************
                            Include header files
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/tree.h>
#include "gridsite.h"
#include "gridsite-gacl.h"

/******************************************************************************
                                Definitions
******************************************************************************/

#define SUCCESS 0
#define FAILURE 1
#define SEARCH_START_QUOTE 1
#define SEARCH_END_QUOTE   2
#define QUOTE_CHAR  '\"'
#define LINESIZE 5000
#define MAXLINES 6000
#define REM "#"
#define DEFAULTOUTPUTFILE "output.gacl"
#define PREFIX_VO        "/VO="
#define PREFIX_GROUP     "/GROUP="
#define PREFIX_SUBGROUP  "/SUBGROUP="
#define PREFIX_ROLE      "/ROLE="
#define PREFIX_CAPS      "/CAPABILITY="
#define PREFIX_SLASHSTAR "/*"
#define PREFIX_STAR      "*"
#define VO_DATA_WHITESPACE_CHARS " \t\n"   

/******************************************************************************
                          Module specific prototypes
******************************************************************************/

void cleanupStatics (void);
void showhelp (void); 
int strclean  (char **s);
int strnclean (char **s, int bufsize);
int handle_parameters (int argc, char **argv);
int handle_gridmapfile (char * gridmapfile,  GACLacl **acl);
int handle_groupmapfile(char * groupmapfile, GACLacl **acl);
int getdnfromvo (char *vo, char ***dn, int *count);
/*
int strsepvoms (const char *voms,
                char **vo, char ***group, char ***subgroup, 
                char ***role, char ***caps,
                int  *ngroup, int *nsubgroup, int *nrole, int *ncaps);
*/
int strsepvoms (const char *voms, char **result, const char *prefix);

int filetest();

/******************************************************************************/

typedef struct {
   char *vo;
   char *dn;
} dnvoms_s;

static char     *gridmapfile     = NULL;
static char     *vomapfile       = NULL;
static char     *groupmapfile    = NULL;
static char     *gaclfile        = NULL;
static int       verboselevel    = 0;
static dnvoms_s *dnvoms          = NULL;
static int       dnvoms_cnt      = 0;
static int       with_star       = 0;
static int       with_slash_star = 0;
static int       warning         = 0;

/******************************************************************************
Function:   main
Description:
    main start of the tool
Parameters:
    argc, argv
    argv[0]: The name of the tool
    argv[?]: parameters
Returns:
    SUCCESS
    FAILURE
******************************************************************************/

int main (int argc,
          char **argv)
{
    GACLacl   *acl = NULL;


    if (handle_parameters(argc, argv) != SUCCESS)
    {
        return FAILURE;
    }
    
    if (verboselevel > 0)
    {
        printf("Parameters successfully gathered\n");
    }

       
    if (filetest() != SUCCESS)
        return FAILURE;


    GACLinit();

    acl = GACLnewAcl();

    if (gridmapfile != NULL)
        handle_gridmapfile(gridmapfile,   &acl);

    if (groupmapfile != NULL)
        handle_groupmapfile(groupmapfile, &acl);

    if (vomapfile != NULL)
        handle_groupmapfile(vomapfile, &acl);

    if ( (gridmapfile  != NULL) || 
         (groupmapfile != NULL) ||
         (vomapfile    != NULL) )
        GACLsaveAcl(gaclfile, acl);

/*
    if ( (gridmapfile  != NULL) || 
         (groupmapfile != NULL) )
        GACLsaveAcl(gaclfile, acl);
*/

    if (verboselevel > 0)
        printf ("Program ended successfully\n");


    GACLfreeAcl(acl);
    free(acl);
    cleanupStatics();

    return SUCCESS;
}

/******************************************************************************
Function:   cleanupStatics
Description:
            Cleans the static variables
Parameters:
Returns:
    SUCCESS
    FAILURE
******************************************************************************/

void cleanupStatics ()
{
    int i = 0;

    if (gridmapfile != NULL)
        free(gridmapfile);

    if (vomapfile != NULL)
        free(vomapfile);

    if (groupmapfile != NULL)
        free(groupmapfile);

    if ( (strncmp(gaclfile, DEFAULTOUTPUTFILE, strlen(DEFAULTOUTPUTFILE)) != 0) &&
         (gaclfile != NULL) )
        free(gaclfile);

    for (i = 0; i < dnvoms_cnt; i++)
    {
        free((dnvoms[i]).vo);
        free((dnvoms[i]).dn);
    }
    free(dnvoms);
}


/******************************************************************************
Function:   handle_gridmapfile
Description:
Parameters:
    gridmapfile
Returns:
    SUCCESS
    FAILURE
******************************************************************************/

int handle_gridmapfile (char * gridmapfile, GACLacl **acl)
{
    GACLcred  *cred      = NULL;
    GACLentry *entry     = NULL;
    FILE      *gridmap   = NULL;
    int        i         = 0;
    char     **lines     = NULL;
    char     **dn        = NULL;
    int        linecount = 0;
    int        dncount   = 0;

    gridmap = fopen(gridmapfile, "r");

    if (gridmap == NULL)
    {
        printf("Error: gridmapfile \"%s\" cannot be opened.\n", gridmapfile);
        return FAILURE;
    }

    /* Dev */
    lines     = NULL; // The first run must be set to NULL
    linecount = 0;    // The first run must start at zero

    do
    {
       linecount++;

       /* Create a line pointer */
       lines = (char **) realloc (lines, (sizeof(char *) * linecount));

       /* Create mem space for a line */
       lines[linecount - 1] = (char *)  malloc (sizeof(char) * LINESIZE);
    }
    while (fgets(lines[linecount - 1], LINESIZE, gridmap) != NULL);


    /* Close file */
    fclose(gridmap);

    /* Clean the lines */
    for (i = 0; i < linecount; i++)
    {
        // No space plz
        lines[i] += strspn(lines[i], VO_DATA_WHITESPACE_CHARS);

        // Clear a empty line
        if (strlen(lines[i]) <= 1)
        {
            continue;
        }

        // Clear lines prefixing a REM block (usually '#')
        if (strncmp(strtok(lines[i], "\"\n"), REM, strlen(REM)) != 0)
        {
            dn = realloc (dn, (sizeof(char *) * ++dncount));
            dn[dncount - 1] = malloc(sizeof(char) * LINESIZE);
            
            dn[dncount - 1] = strtok(lines[i], "\"\n");
        }
    }

    /* DNs */
    if (verboselevel > 1)
    {
        for (i = 0; i < dncount; i++)
        {
            printf("%s\n", dn[i]);
        }
    }

    for (i = 0; i < dncount; i++)
    {
        if (strncmp("/VO", dn[i], strlen("/VO")) != 0)
        {
            cred = GACLnewCred("person");
            GACLaddToCred(cred, "dn", dn[i]);
            
            entry = GACLnewEntry();
            GACLaddCred(entry, cred);

            GACLallowPerm(entry, GACL_PERM_READ);            
            GACLallowPerm(entry, GACL_PERM_WRITE);            
            
            GACLdenyPerm(entry, GACL_PERM_ADMIN);

            GACLaddEntry(*acl, entry);
        }
    }

/*
    // Clean up    
    for (i = 0; i < MAXLINES; i++)
    {
        free(lines[i]);
    }
    free(lines);
*/

/*  Anti Seg. Fault : Probably still in use of GACL lib components.
    for (i = 0; i < dncount; i++)
    {
        free(dn[i]); // Why does Valgrind complain about this free?
    }
    free(dn);
*/

    if (verboselevel > 2)
        printf("End Marker Parsing and cleaning\n");

    // End
    return SUCCESS;
}



/******************************************************************************
Function:   handle_groupmapfile
Description:
Parameters:
    gridmapfile
Returns:
    SUCCESS
    FAILURE
******************************************************************************/

int handle_groupmapfile (char * groupmapfile, GACLacl **acl)
{
    GACLcred  *cred      = NULL;
    GACLentry *entry     = NULL;
    FILE      *groupmap  = NULL;
    int        i         = 0;
    int        t         = 0;
    char     **lines     = NULL;
    char     **vo        = NULL;
    int        linecount = 0;
    int        vocount   = 0;

    /* voms separation data */
    char      *sepvo       = NULL;
    char      *sepgroup    = NULL;
    char      *sepsubgroup = NULL;
    char      *seprole     = NULL;    
    char      *sepcaps     = NULL;
/*
    int        nsubgroup   = 0;
    int        nrole       = 0;
    int        ncaps       = 0; 
    int        ngroup      = 0;
*/   
    /* voms and DNs */
    char     **voms        = 0;
    int        count       = 0;


    groupmap = fopen(groupmapfile, "r");


    if (groupmap == NULL)
    {
        printf("Error: groupmapfile \"%s\" cannot be opened.\n", groupmapfile);
        return FAILURE;
    }

/*
    lines   = (char **) malloc (sizeof(char *) * MAXLINES);
    for (i = 0; i < MAXLINES; i++) { lines[i] = (char *)  malloc (sizeof(char) * LINESIZE); }

    linecount = 0;
    while (fgets(lines[linecount], LINESIZE, groupmap) != NULL) { linecount++; }
*/

    do
    {
       linecount++;

       /* Create a line pointer */
       lines = (char **) realloc (lines, (sizeof(char *) * linecount));

       /* Create mem space for a line */
       lines[linecount - 1] = (char *)  malloc (sizeof(char) * LINESIZE);
    }
    while (fgets(lines[linecount - 1], LINESIZE, groupmap) != NULL);


    /* close file */
    fclose(groupmap);


    /* Clean the lines */
    for (i = 0; i < linecount; i++)
    {
        // No space plz
        lines[i] += strspn(lines[i], VO_DATA_WHITESPACE_CHARS); 

        // Clear a empty line
        if (strlen(lines[i]) <= 1)
        {
            continue;
        }
 
        // Clear lines prefixing a REM block (usually '#')
        if (strncmp(strtok(lines[i], "\"\n"), REM, strlen(REM)) != 0)
        {
            vo = realloc (vo, (sizeof(char *) * ++vocount));
            vo[vocount - 1] = malloc(sizeof(char) * LINESIZE);
            
            vo[vocount - 1] = strtok(lines[i], "\"\n");
        }
    }


    /* VO-GROUP-ROLEs */
    if (verboselevel > 0)   
    {
        for (i = 0; i < vocount; i++)
        {
            printf("%s\n", vo[i]);
        }
    }

    sepvo       = (char *) malloc(sizeof(char) * LINESIZE);
    sepgroup    = (char *) malloc(sizeof(char) * LINESIZE);
    sepsubgroup = (char *) malloc(sizeof(char) * LINESIZE);
    seprole     = (char *) malloc(sizeof(char) * LINESIZE);
    sepcaps     = (char *) malloc(sizeof(char) * LINESIZE);


    /* VO-GROUP-ROLE parsing */
    for (i = 0; i < vocount; i++)
    {
        if (strncmp("/VO", vo[i], strlen("/VO")) == 0)
        {

            strnclean (&sepvo, LINESIZE);
            strnclean (&sepgroup, LINESIZE);
            strnclean (&sepsubgroup, LINESIZE);
            strnclean (&seprole, LINESIZE);
            strnclean (&sepcaps, LINESIZE);


            strsepvoms(vo[i], &sepvo,       "/VO=");
            strsepvoms(vo[i], &sepgroup,    "/GROUP=");
            strsepvoms(vo[i], &sepsubgroup, "/SUBGROUP=");
            strsepvoms(vo[i], &seprole,     "/ROLE=");
            strsepvoms(vo[i], &sepcaps,     "/CAPABILITY=");

            if (voms != NULL)
            {
                for (t = 0; t < count; t++)
                {
                    strclean(&(voms[t]));
                    free(voms[t]);
                }
                free(voms);
                voms = NULL;
            }

            if (getdnfromvo (sepvo, &voms, &count) == SUCCESS)
            {
                for (t = 0; t < count; t++)
                {
                    cred = GACLnewCred("voms-cred");

                    GACLaddToCred(cred, "voms", voms[t]);
            
                    if (strlen(sepvo) > 0)
                        GACLaddToCred(cred, "vo", sepvo);
                    if (strlen(sepgroup) > 0)
                        GACLaddToCred(cred, "group", sepgroup);
                    if (strlen(sepsubgroup) > 0)
                        GACLaddToCred(cred, "subgroup", sepsubgroup);
                    if (strlen(seprole) > 0)
                        GACLaddToCred(cred, "role", seprole);
                    if (strlen(sepcaps) > 0)
                        GACLaddToCred(cred, "capability", sepcaps);
                
                    entry = GACLnewEntry();
                    GACLaddCred(entry, cred);
                    GACLallowPerm(entry, GACL_PERM_READ);            
                    GACLallowPerm(entry, GACL_PERM_WRITE);            
                
                    GACLdenyPerm(entry, GACL_PERM_ADMIN);
    
                    GACLaddEntry(*acl, entry);
                }        
            }
            else
            {
                if (!warning)
                    printf("Warning: This entry with VO '%s' contains no data about the DN of the VOMS server.\n", sepvo);

                cred = GACLnewCred("voms-cred");
    
//                GACLaddToCred(cred, "voms", vo[i]);  // No knowledge yet about which VOMS server should be inserted here
//                GACLaddToCred(cred, "voms", "");
    
                if (strlen(sepvo) > 0)
                    GACLaddToCred(cred, "vo", sepvo);
                if (strlen(sepgroup) > 0)
                    GACLaddToCred(cred, "group", sepgroup);
                if (strlen(sepsubgroup) > 0)
                    GACLaddToCred(cred, "subgroup", sepsubgroup);
                if (strlen(seprole) > 0)
                    GACLaddToCred(cred, "role", seprole);
                if (strlen(sepcaps) > 0)
                    GACLaddToCred(cred, "capability", sepcaps);
                
                entry = GACLnewEntry();
                GACLaddCred(entry, cred);
                GACLallowPerm(entry, GACL_PERM_READ);            
                GACLallowPerm(entry, GACL_PERM_WRITE);            
                
                GACLdenyPerm(entry, GACL_PERM_LIST);            
                GACLdenyPerm(entry, GACL_PERM_ADMIN);
    
                GACLaddEntry(*acl, entry);
            }

            strnclean (&sepvo, LINESIZE);
            strnclean (&sepgroup, LINESIZE);
            strnclean (&sepsubgroup, LINESIZE);
            strnclean (&seprole, LINESIZE);
            strnclean (&sepcaps, LINESIZE);
        }
    }

    free(sepvo);
    free(sepgroup);
    free(sepsubgroup);
    free(seprole);
    free(sepcaps);

/*
    // Clean up    
    for (i = 0; i < MAXLINES; i++)
    {
        free(lines[i]);
    }
    free(lines);
*/


/*  For details see handle_gridmapfile function
    for (i = 0; i < vocount; i++)
    {
        free(vo[i]);
    }
    free(vo);
*/


    if (verboselevel > 2)
        printf("End Marker Parsing and cleaning\n");

    return SUCCESS;
}


/******************************************************************************
Function:    strsepvoms
Description: Separates a string of voms credentials into different 
             parameters starting with the 'prefix'
Parameters:
     input : char *voms,
             const char *prefix 
     output: char ** result 
Returns:
    SUCCESS
    FAILURE
BUGS:
    -    It doesn't get the second- or subgroup like: .../GROUP=grp1/grp2
         This will be fixed.
    -    Multiple Groups, Roles, Capabilities cannot yet be gathered and transformed
         into GACL

******************************************************************************/

int strsepvoms (const char *voms, char **result, const char *prefix)
{
    int i = 0;
    char *tmpvoms   = NULL;
    char *tmpresult = NULL;
    char *oldtmpvoms = NULL;

    tmpvoms   = strdup(voms);
    tmpresult = (char *) malloc(sizeof(char) * LINESIZE);
    strnclean(&tmpresult, LINESIZE);

    // debug
    oldtmpvoms = tmpvoms;

    do
    {
        // Determine case-insensitive if the prefix is at the beginning of the string
        if (strncasecmp(tmpvoms, prefix, strlen(prefix)) == 0)
        {
            if (verboselevel > 3)
                printf("Prefix \"%s\" found, stepping over prefix string\n", prefix);

            // Stepping over prefix
            for (i = 0; i < strlen(prefix); i++)
            {
                tmpvoms++;
            }

            // Determining the length of the found string by searching for an end character
            i = 0;
            while (tmpvoms[i] != '\0')
            {
                if (strlen(tmpvoms) - i >= (strlen(PREFIX_VO)))
                {
                    if (strncmp(PREFIX_VO, &tmpvoms[i], strlen(PREFIX_VO)) == 0)
                    {
                        // You have found the end character of the wanted substring
                        break;
                    }
                }

                if (strlen(tmpvoms) - i >= (strlen(PREFIX_GROUP)))
                {
                    if (strncmp(PREFIX_GROUP, &tmpvoms[i], strlen(PREFIX_GROUP)) == 0)
                    {
                        break;
                    }
                }

                if (strlen(tmpvoms) - i >= (strlen(PREFIX_SUBGROUP)))
                {
                    if (strncmp(PREFIX_SUBGROUP, &tmpvoms[i], strlen(PREFIX_SUBGROUP)) == 0)
                    {
                        break;
                    }
                }

                if (strlen(tmpvoms) - i >= (strlen(PREFIX_ROLE)))
                {
                    if (strncmp(PREFIX_ROLE, &tmpvoms[i], strlen(PREFIX_ROLE)) == 0)
                    {
                        break;
                    }
                }

                if (strlen(tmpvoms) - i >= (strlen(PREFIX_CAPS)))
                {
                    if (strncmp(PREFIX_CAPS, &tmpvoms[i], strlen(PREFIX_CAPS)) == 0)
                    {
                        break;
                    }
                }

                // if the cmdline parameter -with_slash_star is set it will add the star behind the group name
                // like    VOMS :  /GROUP=fred/*  =>  GACL :  <GROUP>fred/*</GROUP>
                // Default VOMS :  /GROUP=fred/*  =>  GACL :  <GROUP>fred</GROUP>
                if (!with_slash_star)
                {
                    if (strlen(tmpvoms) - i >= (strlen(PREFIX_SLASHSTAR)))
                    {
                        if (strncmp(PREFIX_SLASHSTAR, &tmpvoms[i], strlen(PREFIX_SLASHSTAR)) == 0)
                        {
                            break;
                        }
                    }
                }

                // if the cmdline parameter -with_star is set it will add the star behind the group name
                // like    VOMS :  /GROUP=fred*  =>  GACL :  <GROUP>fred*</GROUP>
                // Default VOMS :  /GROUP=fred*  =>  GACL :  <GROUP>fred</GROUP>
                if (!with_star)
                {
                    if (strlen(tmpvoms) - i >= (strlen(PREFIX_STAR)))
                    {
                        if ( (strncmp(PREFIX_STAR, &tmpvoms[i], strlen(PREFIX_STAR)) == 0) && 
                             (i > 0) && 
                             (strncmp(PREFIX_SLASHSTAR, &tmpvoms[i - 1], strlen(PREFIX_SLASHSTAR)) != 0) )
                        {
                            break;
                        }
                    }
                }

                // If none of the above aprove ... keep on counting
                i++;
            }

            // Determining the length of the found string by searching for an end character
//            i = 0;
//            while ((tmpvoms[i] != '/') && (tmpvoms[i] != '\0') && (tmpvoms[i] != '*'))
//            {
//                i++;
//            }

            // Copy found string to a result of determined length
            strncpy(tmpresult, tmpvoms, i);
            
            if (verboselevel > 3)
            {
                printf ("\ttmpresult = \"%s\"\n", strlen(tmpresult) ? tmpresult : "(null)");
            }

            // Stop do-while loop because we already found something and 
            // we can only find the first occurence of the prefix in the string
            break;
        }
        else
        {
            strnclean(&tmpresult, LINESIZE);
        }

        // Stepping thru a line to be parsed
        strcpy(tmpvoms, &tmpvoms[1]); // Step forwards
    }
    while (i < strlen(tmpvoms));

    strncpy(*result, tmpresult, strlen(tmpresult));

    free(tmpresult);
    free(oldtmpvoms);

    return SUCCESS;

}

/******************************************************************************
Function:   showhelp
Description:
Parameters:
Returns:
******************************************************************************/
void showhelp () 
{
    printf("\n");
    printf("\t-gridmapfile  <filepath>  -  Handles a 'gridmapfile' or 'grid-mapfile' as \n");
    printf("\t                             we use it with for local- and poolaccount\n");
    printf("\t-groupmapfile <filepath>  -  Handles a 'groupmapfile' with VO-GROUP-ROLEs and\n");
    printf("\t                             map-information for the local- and poolgroups\n");
//    printf("\t-vomapfile    <filepath>  -  Handles a 'vomapfile' (this information is sometimes\n");
//    printf("\t                             at the end of a 'gridmapfile') containing\n");
//    printf("\t                             VO-GROUP-ROLE map-information to poolaccounts\n");
    printf("\t-gaclfile     <filepath>  -  Assigned output file (default: '%s')\n", DEFAULTOUTPUTFILE);
    printf("\t-vomsdn  <VO> <voms dn>   -  Overrides the DN with <voms dn> for every entry that is\n");
    printf("\t                             associated with the VO <VO> of the voms server\n");
    printf("\t                             NOTE: a VO can have multiple voms servers\n");
    printf("\t-v[v[v[v]]]               -  Extra diagnostic information\n");
    printf("\t                             NOTE: a verboselevel above 1 will trigger a file check\n");
    printf("\t                                   a verboselevel above 3 will give extra parsing information\n");
    printf("\t-with_star                -  When this parameter is set, the GACL file will include the wildcard\n");
    printf("\t                             in the credential part of the ACL Entry.\n");
    printf("\t                             NOTE: this is for all known VOMS credential types: VO, GROUP, SUBGROUP, ROLE or CAPABILITY\n");
    printf("\t-with_slash_star          -  When this parameter is set, the GACL file will include the wildcard\n");
    printf("\t                             in the credential part of the ACL Entry.\n");
    printf("\t                             NOTE: this is for all known VOMS credential types: VO, GROUP, SUBGROUP, ROLE or CAPABILITY\n");
    printf("\t-nowarn                   -  Skips the warning messages if they do not indicate an error.\n");
    printf("\n");
}



/******************************************************************************
Function:   handle_parameters
Description:
Parameters:
    argc, argv
Returns:
    SUCCESS
    FAILURE
******************************************************************************/

int handle_parameters (int argc, char **argv)
{
    int i;

    /*
     * Atleast one parameter must be present to do 'something'
     */

    if (!(argc > 1))
    {
        printf("Error: no parameters found, nothing to do...\n");   
        showhelp();

        return FAILURE;
    }

    for (i = 1; i < argc; i++)
    {
        if ( (strcmp(argv[i], "-vomsdn") == 0) && (i + 2< argc))
        {
            if ((argv[i + 1] != NULL) && 
                (strlen(argv[i + 1]) > 0) &&
                (argv[i + 2] != NULL) &&
                (strlen(argv[i + 2]) > 0))
            {
                 dnvoms = realloc(dnvoms, sizeof(dnvoms_s) * ++dnvoms_cnt);
                 (dnvoms[dnvoms_cnt - 1]).vo = strdup(argv[i + 1]);
                 (dnvoms[dnvoms_cnt - 1]).dn = strdup(argv[i + 2]);

            }
            i += 2;;
            
        }

        else if ( (strcmp(argv[i], "-gridmapfile") == 0) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 gridmapfile = strdup(argv[i + 1]);
            }
            i++;
        }

        else if ( (strcmp(argv[i], "-groupmapfile") == 0) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 groupmapfile = strdup(argv[i + 1]);
            }
            i++;
        }

        else if ( (strcmp(argv[i], "-vomapfile") == 0) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 vomapfile = strdup(argv[i + 1]);
            }
            i++;
        }

        else if ( (strcmp(argv[i], "-gaclfile") == 0) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 gaclfile = strdup(argv[i + 1]);
            }
            i++;
        }

        else if (strncmp(argv[i], "-v", 2) == 0)
        {
            int j = 0;

            verboselevel = 0;
            while (j < strlen(argv[i]))
            {
                if (((argv[i])[j]) == 'v')
                    verboselevel++;

                j++;
            }
        }

        else if ( (strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "-help") == 0) )
        {
            showhelp();
        }

        else if ( (strcmp(argv[i], "-with_star") == 0) || (strcmp(argv[i], "-with-star") == 0) )
        {
            with_star = 1;
        }

        else if ( (strcmp(argv[i], "-with_slash_star") == 0) || (strcmp(argv[i], "-with-slash-star") == 0) )
        {
            with_slash_star = 1;
        }

        else if (strcmp(argv[i], "-nowarn") == 0) 
        {
            warning = 1;
        }

        else
        {
            printf("\n\tError: something is wrong with the parameters.\n");
              printf("\t       The error has been detected on \"%s\"\n", argv[i]);
              printf("\t       The last two parameters are : \"%s\" and \"%s\"\n\n", argv[i - 1], argv[i]);
            showhelp();
            return FAILURE;
        }
    }

    if (verboselevel > 0)
    {
       printf("verbose level = %d\n", verboselevel);
    }

    if (gaclfile == NULL)
    {
        gaclfile = DEFAULTOUTPUTFILE;
    }

    if (verboselevel > 2)
    {
       printf("Files initialized as:\n");
       printf("-gridmapfile  = %s\n", gridmapfile);
       printf("-groupmapfile = %s\n", groupmapfile);
       printf("-vomapfile    = %s\n", vomapfile);
       printf("-gaclfile     = %s\n", gaclfile);
    }
    else if ( (verboselevel > 0) && (verboselevel <= 2) )
    {
        printf("Output is set to: %s\n", gaclfile);
    } 

    if (verboselevel > 0)
    {
        for (i = 0; i < dnvoms_cnt; i++)
            printf("<VO> : %s \tis associated with \t<DN> : %s\n", (dnvoms[i]).vo, (dnvoms[i]).dn);
    }

    return SUCCESS;
}

/******************************************************************************
Function:   filetest
Description:
Parameters:
    argc, argv
Returns:
    SUCCESS
    FAILURE
******************************************************************************/

int filetest()
{
    FILE *test = NULL;
	
    if (gridmapfile != NULL)
    {
        if ((test = fopen(gridmapfile, "r")) != NULL)
        {
            if (verboselevel > 0)
                printf("Success: gridmapfile \"%s\" readable\n", gridmapfile);

            fclose(test);
        }
        else
        {
            printf("Error: Can't open gridmapfile \"%s\"\n", gridmapfile);
            return FAILURE;
        }
    }

    if (groupmapfile != NULL)
    {
        if ((test = fopen(groupmapfile, "r")) != NULL)
        {
            if (verboselevel > 0)
                printf("Success: groupmapfile \"%s\" readable\n", groupmapfile);

            fclose(test);
        }
        else
        {
            printf("Error: Can't open groupmapfile \"%s\"\n", groupmapfile);
            return FAILURE;
        }
    }

    if (vomapfile != NULL)
    {
        if ((test = fopen(vomapfile, "r")) != NULL)
        {
            if (verboselevel > 0)
                printf("Success: vomapfile \"%s\" readable\n", vomapfile);

            fclose(test);
        }
        else
        {
            printf("Error: Can't open vomapfile \"%s\"\n", vomapfile);
            return FAILURE;
        }
    }

    if (verboselevel > 1)
    {
        if (gaclfile != NULL)
        {
            if ((test = fopen(gaclfile, "r")) != NULL)
                fclose(test);
            else
            {
                printf("Warning: Can't open gaclfile \"%s\"\nA new gaclfile will be used.", gaclfile);
            }
        }
    }

    return SUCCESS;
}




/******************************************************************************
Function:    strnclean
Description: cleans a dirty string, forces '\0' on the bufsize of **s
Parameters:
     input  : char **s
     output : char **s (cleaned string and NULLified)
Returns:
    SUCCESS
    FAILURE

******************************************************************************/
int strnclean (char **s, int bufsize)
{
    int i = 0;

    if (s == NULL)
        return FAILURE;

    if (*s == NULL)
        return FAILURE;

    for (i = 0; i < bufsize; i++)
    {
        (*s)[i] = '\0';
    }

    return SUCCESS;
}


/******************************************************************************
Function:    strclean
Description: cleans a dirty string, forces '\0' on the strlen() of **s
Parameters:
     input  : char **s
     output : char **s (cleaned string and NULLified)
Returns:
    SUCCESS
    FAILURE

******************************************************************************/
int strclean (char **s)
{
    if (s == NULL)
        return FAILURE;

    if (*s == NULL)
        return FAILURE;

    if (strlen(*s) > 0)
    {
        int i = 0;
        
        for (i = 0; i < strlen(*s); i++)
            (*s)[i] = '\0';
    }

    return SUCCESS;
}
/*****************************************************************************/


/******************************************************************************
Function:    getdnfromvo
Description: searches for the VO in the dnvoms array. The function assigns a DN of
             a VOMS server to a VO.
Parameters:
     input  : char *vo
     output : char ***dn, int *count
Returns:
    SUCCESS
    FAILURE
 
******************************************************************************/
int getdnfromvo (char *vo, char ***dn, int *count)
{
    int i;
    int resultvalue = FAILURE;
    char **tmp = NULL;

    if (*dn != NULL)
        return FAILURE;

    if (dnvoms_cnt < 1)
        return FAILURE;

    *count = 0;

    for (i = 0; i < dnvoms_cnt; i++)
    {
         if (strncmp((dnvoms[i]).vo, vo, strlen((dnvoms[i]).vo)) == 0)
         {
             tmp = (char **) realloc(*dn, ++(*count) * sizeof(char *));

             tmp[*count - 1] = strdup((dnvoms[i]).dn);
//             *(dn[*count - 1]) = strdup((dnvoms[i]).dn);

             *dn = tmp;
             resultvalue = SUCCESS;
         }
    }
    return resultvalue;
}
