/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
    \page lcas_voms.mod voms plugin

    \section vomssyn SYNOPSIS
        \b lcas_voms.mod -vomsdir \<vomsdir\> -certdir \<certdir\>
           -authfile \<authorization file\>  [-authformat \<format of the authorization file\>]

    \section vomsdescr DESCRIPTION

        This plugin forms the link between the VOMS data found in the user grid credential (X509
        certificate) and the lcas system.
        It will retrieve the VOMS data by using the VOMS API. The VOMS data will be checked against either 
        a (simple) gridmap style file, a GACL-file or an XACML-file in order for the user job to be authorized on the site.

    \section vomsoptions OPTIONS
    \subsection vomsoptie1 -VOMSDIR <vomsdir>
        See \ref vomsoptie2 "-vomsdir"

    \subsection vomsoptie2 -vomsdir <vomsdir>
        This is the directory which contains the certificates of the VOMS servers

    \subsection vomsoptie3 -CERTDIR <certdir>
        See \ref vomsoptie4 "-certdir"
 
    \subsection vomsoptie4 -certdir <certdir>
        This is the directory which contains the CA certificates

    \subsection vomsoptie5 -authfile <authorization file>
        In this file the authorization/access control based on VOMS information is specified.
        The format of this file is 'simple' (gridmap style), 'gacl' or 'xacml', which can be specified
        explicitly with the option \ref vomsoptie6 "-authformat" or will be derived form the suffix
        of the authorization file (\c .gacl and \c .xacml for 'gacl' and 'xacml' formats, otherwise 'simple').

    \subsection vomsoptie6 -authformat <format of the authorization file>
        Format of the autorization file, values: \c gacl/GACL, \c xacml/XACML or \c simple.

    \subsection vomsoptie7 -gacl_use_voms_dn [yes|no|always]
        GACL specific.
        This option specifies if the voms DN, found in the user certificate, should be included in
        the user gacl credential. Default is 'yes'. The following arguments are recognized:
        \li \c yes : For each VO-GROUP-ROLE combination found in the user
        certificate two gacl credentials are created:
        one \e with and one \e without the voms DN. In this way the user is also authorized if in
        the gacl in the authorization file the voms DN is not included (better if it is, though).
        \li \c always : For each VO-GROUP-ROLE combination found in the user
        certificate only a gacl credential is created \e with the voms DN.
        \li \c no : For each VO-GROUP-ROLE combination found in the
        usercertificate a gacl credential is created \e without the voms DN.

    \subsection vomsoptie8 -use_user_dn
        If this option is set also user proxies without voms information will be processed. If the
        user dn of the proxy is present in the gacl or gridmapfile, the user is authorized by this plugin.
        

\section vomsReturnvalue RETURN VALUES
        \li LCAS_MOD_SUCCESS : Success
        \li LCAS_MOD_FAIL    : Failure
 
 
\section vomsErrors ERRORS
        See bugzilla for known errors (http://marianne.in2p3.fr/datagrid/bugzilla/)
 
\section vomsSeeAlso SEE ALSO
        \ref lcas_userallow.mod "lcas_userallow.mod",
        \ref lcas_userban.mod "lcas_userban.mod",
        \ref lcas_timeslots.mod "lcas_timeslots.mod",
*/

/*!
    \file   lcas_voms.c
    \brief  Interface to the LCAS plugins
    \author Martijn Steenbakkers for the EU DataGrid.

    This file contains the code for the voms plugin (extracts the VOMS info from the
    certificate). The interface consists of the following functions:
    -# plugin_initialize()
    -# plugin_confirm_authorization()
    -# plugin_terminate()
*/

/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcas_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>
#include <openssl/x509.h>
#include <errno.h>

#include <gssapi.h>

#include <lcas/lcas_modules.h>

#include "lcas_voms_utils.h"
#include "lcas_vo_data.h"
#include "lcas_gridlist.h"

#include <libxml/tree.h>
#include <gridsite.h>
#include <gridsite-gacl.h>

#include <voms/voms_apic.h>
#include <globus_gss_assist.h>

#undef TESTBIO
#ifdef TESTBIO
#    include <openssl/pem.h>
#    include <openssl/bio.h>
#endif


/******************************************************************************
                                Definitions
******************************************************************************/

#define VOMS_BUFFER_SIZE 1024

/******************************************************************************
                               Type definitions
******************************************************************************/
/*!
    \enum authformat_e
    \brief This enumeration type gives the different plugin symbol/function types.
    \internal
*/
typedef enum authformat_e
{
    NO_FORMAT,     /*!< this value corresponds to no format at all */
    SIMPLE_FORMAT, /*!< this value corresponds to the 'simple' authorization file format */
    GACL_FORMAT,   /*!< this value corresponds to the 'gacl' authorization file format */
    XACML_FORMAT,  /*!< this value corresponds to the 'xacml' authorization file format */
}
authformat_t;

typedef enum gacl_use_voms_dn_e
{
    ALWAYS_USE_VOMS_DN,  /*!< always include the VOMS dn in the gacl user credential */
    USE_VOMS_DN,         /*!< create a gacl user credential with and without the VOMS dn */
    DONT_USE_VOMS_DN,    /*!< do not include the VOMS dn in the gacl user credential */
}
gacl_use_voms_dn_t;

typedef enum authz_cred_e
{
    USE_GSS,    /*!< Use gss_cred_id_t for the authorization decision */
    USE_X509,   /*!< Use the user's X509 proxy cert + chain for the authorization decision */
}
authz_cred_t;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/
static int _plugin_confirm_authorization(authz_cred_t, lcas_request_t, lcas_cred_id_t, X509 *, STACK_OF(X509) *);
static void print_vomsdata(struct vomsdata *);
static void print_vomsdata(struct vomsdata *);
static int  lcas_check_gacl(GACLuser *, char *);
static int  lcas_gacl_add_dn(GACLuser **, char *);
static int  lcas_gacl_add_vomsdata(GACLuser **, lcas_vo_data_t *, char *);
void        lcas_print_x509_to_string (X509 *, char *);

#ifdef TESTBIO
static STACK_OF(X509) *load_chain(char *certfile);
static char *retmsg[] = { "VERR_NONE", "VERR_NOSOCKET", "VERR_NOIDENT", "VERR_COMM", 
                          "VERR_PARAM", "VERR_NOEXT", "VERR_NOINIT",
                          "VERR_TIME", "VERR_IDCHECK", "VERR_EXTRAINFO",
                          "VERR_FORMAT", "VERR_NODATA", "VERR_PARSE",
                          "VERR_DIR", "VERR_SIGN", "VERR_SERVER", 
                          "VERR_MEM", "VERR_VERIFY", "VERR_IDENT",
                          "VERR_TYPE", "VERR_ORDER" };
#endif

/******************************************************************************
                       Define module specific variables
******************************************************************************/


static authformat_t authformat       = NO_FORMAT;
static char *       authfile         = NULL;
static char *       certdir          = NULL;
static char *       vomsdir          = NULL;
static char         voms_buffer[VOMS_BUFFER_SIZE];
static authformat_t gacl_use_voms_dn = USE_VOMS_DN;
static int          use_user_dn      = 0;
static int          debug_certs      = 0;
static char *       debug_printfile  = NULL;


#ifdef TESTBIO
static STACK_OF(X509) *load_chain(char *certfile)
{
    STACK_OF(X509_INFO) *sk=NULL;
    STACK_OF(X509) *stack=NULL, *ret=NULL;
    BIO *in=NULL;
    X509_INFO *xi;
    int first = 1;

    if(!(stack = sk_X509_new_null()))
    {
        printf("%s: memory allocation failure\n", logstr);
        goto end;
    }

    if(!(in=BIO_new_file(certfile, "r")))
    {
        printf("%s: error opening the file, %s\n", logstr,certfile);
        goto end;
    }

    /* This loads from a file, a stack of x509/crl/pkey sets */
    if(!(sk=PEM_X509_INFO_read_bio(in,NULL,NULL,NULL)))
    {
        printf("%s: error reading the file, %s\n", logstr,certfile);
        goto end;
    }

    /* scan over it and pull out the certs */
    while (sk_X509_INFO_num(sk))
    {
        /* skip first cert */
        if (first)
        {
            first = 0;
            continue;
        }
        xi=sk_X509_INFO_shift(sk);
        if (xi->x509 != NULL)
        {
            sk_X509_push(stack,xi->x509);
            xi->x509=NULL;
        }
        X509_INFO_free(xi);
    }
    if(!sk_X509_num(stack))
    {
        printf("%s: no certificates in file, %s\n", logstr,certfile);
        sk_X509_free(stack);
        goto end;
    }
    ret=stack;
end:
    BIO_free(in);
    sk_X509_INFO_free(sk);
    return(ret);
}
#endif



#undef EXPORT_CREDENTIAL
#if EXPORT_CREDENTIAL
/******************************************************************************
Function:   exportGSSCredentials
Description:
    export the GSS credential to a file
Parameters:
    cred:                GSS credentials
Returns:
    0  : succes
    !0 : failure
******************************************************************************/
int exportCredentials (gss_cred_id_t cred)
{
    char * logstr = "\tlcas_plugin_voms-exportCredentials()";

    if (cred)
    {
        gss_buffer_desc                 deleg_proxy_filename;
        OM_uint32         major_status = 0;
        OM_uint32         minor_status = 0;

        major_status = gss_export_cred(&minor_status,
                                       cred,
                                       NULL,
                                       1,
                                       &deleg_proxy_filename);

        if (major_status == GSS_S_COMPLETE)
        {
            char *                      cp;

            lcas_log_debug(1,"%s: deleg_proxy_filename.value: %s\n", logstr,
                               deleg_proxy_filename.value);
            cp = strchr((char *)deleg_proxy_filename.value, '=');
            *cp = '\0';
            cp++;
            setenv((char *)deleg_proxy_filename.value, cp, 1);
            free(deleg_proxy_filename.value);
        }
        else
        {
            char *                      error_str = NULL;
            globus_object_t *           error_obj;

            error_obj = globus_error_get((globus_result_t) minor_status);

            error_str = globus_error_print_chain(error_obj);
            lcas_log(0,"%s: Error, gss_export_cred(): %s\n", logstr,error_str);
            return 1;
        }
    }
    else
    {
        lcas_log(0, "%s: Error finding credential struct to export\n", logstr);
        return 1;
    }
}
#endif /* EXPORT_CREDENTIAL */


/******************************************************************************
Function:   extractX509fromGSS
Description:
    extract the X509 credentials from the GSS container
Parameters:
    cred:                GSS credentials
    px509_cred_gss:      X509 user (proxy) cert
    px509_chain_gss:     X509 proxy chain
Returns:
    0  : succes
    !0 : failure
******************************************************************************/
int extractX509fromGSS (gss_cred_id_t    cred,
                        X509 *           px509_cred_gss,
                        STACK_OF(X509) * px509_chain_gss)
{
    char * logstr = "\tlcas_plugin_voms-extractX509fromGSS()";

    /* Need to find credentials */
    if (cred != GSS_C_NO_CREDENTIAL)
    {
        if ( ( px509_cred_gss = lcas_cred_to_x509(cred) ) )
        {
            lcas_log_debug(1,"%s: found X509 struct inside gss credential\n", logstr);
            lcas_log_debug(5,"%s: just for kicks: X509->name %s\n", logstr,px509_cred_gss->name);
        }
        else
        {
            lcas_log(0,"%s: could not get X509 cred (exit voms)!\n", logstr);
            return 1;
        }
        if ( ( px509_chain_gss = lcas_cred_to_x509_chain(cred) ) )
        {
            lcas_log_debug(1,"%s: found X509 chain inside gss credential\n", logstr);
        }
        else
        {
            lcas_log(0,"%s: could not get X509 chain (exit voms)!\n", logstr);
            return 1;
        }
    }
    else
    {
        lcas_log(0,"%s: GSS credential extraction failed: No data structure found here\n", logstr);
        return 1;
    }

    return 0;
}






/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
    LCAS_MOD_NOFILE  : db file not found (will halt LCAS initialization)
******************************************************************************/
int plugin_initialize(
        int argc,
        char ** argv
)
{
    struct stat  buf;
    char *       logstr = "\tlcas_plugin_voms-plugin_initialize()";
    char *       authfilecopy = NULL;
    char *       authfilebase = NULL;
    int          i;
 
    lcas_log_debug(1,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
       lcas_log_debug(2,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }
 
    /*
     * CERTDIR = The directory which contains the CA certificates
     * VOMSDIR = The directory which contains the certificates of the VOMS servers 
     */
 
    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */
    for (i = 1; i < argc; i++)
    {
        if ( ((strcmp(argv[i], "-vomsdir") == 0) ||
              (strcmp(argv[i], "-VOMSDIR") == 0))
             && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 vomsdir = strdup(argv[i + 1]);
            }
            i++;
        }
        else if ( ((strcmp(argv[i], "-certdir") == 0) ||
                   (strcmp(argv[i], "-CERTDIR") == 0))
                   && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 certdir = strdup(argv[i + 1]);
            }
            i++;
        }
        else if ((strcmp(argv[i], "-authfile") == 0)
                   && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 authfile = strdup(argv[i + 1]);
            }
            i++;
        }
        else if ((strcmp(argv[i], "-authformat") == 0)
                   && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 if ( (strcmp(argv[i + 1], "gacl") == 0) || (strcmp(argv[i + 1], "GACL") == 0) )
                 {
                     authformat = GACL_FORMAT;
                 }
                 else if ( (strcmp(argv[i + 1], "xacml") == 0) || (strcmp(argv[i + 1], "XACML") == 0) )
                 {
                     authformat = XACML_FORMAT;
                 }
                 else if ( (strcmp(argv[i + 1], "simple") == 0) || (strcmp(argv[i + 1], "SIMPLE") == 0) )
                 {
                     authformat = SIMPLE_FORMAT;
                 }
                 else
                 {
                     lcas_log(0,"%s: use  \"simple/SIMPLE\", \"gacl/GACL\" or \"xacml/XACML\" for option %s\n",
                                logstr, argv[i]);
                     goto fail_voms_init;
                 }
            }
            i++;
        }
        else if (strcmp(argv[i], "-use_user_dn") == 0)
        {
            use_user_dn = 1;
        }
        else if ( (strcmp(argv[i], "-debug_certs") == 0)
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 debug_certs = 1;
                 debug_printfile = strdup(argv[i + 1]);
            }
            i++;
        }
        else if ((strcmp(argv[i], "-gacl_use_voms_dn") == 0)
                   && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                 if ( (strcmp(argv[i + 1], "yes") == 0) || (strcmp(argv[i + 1], "YES") == 0) )
                 {
                     gacl_use_voms_dn = USE_VOMS_DN;
                 }
                 else if ( (strcmp(argv[i + 1], "no") == 0) || (strcmp(argv[i + 1], "NO") == 0) )
                 {
                     gacl_use_voms_dn = DONT_USE_VOMS_DN;
                 }
                 else if ( (strcmp(argv[i + 1], "always") == 0) || (strcmp(argv[i + 1], "ALWAYS") == 0) )
                 {
                     gacl_use_voms_dn = ALWAYS_USE_VOMS_DN;
                 }
                 else
                 {
                     lcas_log(0,"%s: use  \"yes/YES\", \"no/NO\" or \"always/ALWAYS\" for option %s\n",
                                logstr, argv[i]);
                     goto fail_voms_init;
                 }
            }
            i++;
        }
        else
        {
            lcas_log(0,"%s: Error in initialization parameter: %s (failure)\n", logstr,
                       argv[i]);
            return LCAS_MOD_FAIL;
        }
    }

    /* Check if authfile is filled */
    if (authfile == NULL)
    {
        lcas_log(0, "%s: authorization file is not specified (failure)\n", logstr);
        goto fail_voms_init;
    }
    if ((stat(authfile, &buf)) != 0)
    {
        lcas_log(0, "%s: unable to get stat of authorization file: %s\n", logstr, authfile);
        lcas_log(0, "%s: %s\n", logstr, strerror(errno));
        goto fail_voms_init;
    }
    if ( !(S_ISREG(buf.st_mode)))
    {
        lcas_log(0, "%s: specified authorization file (%s) is not a regular file (failure)\n",
                   logstr, authfile);
        goto fail_voms_init;
    }

    /* If necessary, try to establish the authorization file format from suffix */
    if (authformat == NO_FORMAT)
    {
        authfilecopy = strdup(authfile);
        authfilebase = basename(authfilecopy);
        if (strstr(authfilebase, ".gacl") != NULL)
        {
            authformat = GACL_FORMAT;
        }
        else if (strstr(authfilebase, ".xacml") != NULL)
        {
            authformat = XACML_FORMAT;
        }
        else
        {
            authformat = SIMPLE_FORMAT;
        }
        if (authfilecopy)
        {
            free(authfilecopy);
            authfilecopy = NULL;
        }
    }
    lcas_log_debug(1, "%s: authorization file: %s, format: %d\n", logstr, authfile, authformat);

    /* success */
// success_voms_init:
    return LCAS_MOD_SUCCESS;

    /* failure */
 fail_voms_init:
    return LCAS_MOD_FAIL;
} 


/******************************************************************************
Function:   plugin_confirm_authorization
Description:
    Ask for authorization by passing RSL and user credential
Parameters:
    request:   RSL request
    user_cred: user credential
Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
******************************************************************************/
int
plugin_confirm_authorization(lcas_request_t request, lcas_cred_id_t lcas_cred)
{
    return _plugin_confirm_authorization(USE_GSS, request, lcas_cred, NULL, NULL);
}

/******************************************************************************
Function:   plugin_confirm_authorization_from_x509
Description:
    Ask for authorization by passing RSL, the X509 proxy + chain, and
    (optionally) the lcas credential
Parameters:
    request:            RSL request
    user_cred:          user credential
    px509_cred_in:      X509 user (proxy) cert
    px509_chain_in:     X509 proxy chain
Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
******************************************************************************/
int
plugin_confirm_authorization_from_x509(
    lcas_request_t request,
    lcas_cred_id_t lcas_cred,
    X509 * px509_cred_in,
    STACK_OF(X509) * px509_chain_in
)
{
    return _plugin_confirm_authorization(USE_X509, request, lcas_cred, px509_cred_in, px509_chain_in);
}



static int
_plugin_confirm_authorization(
    authz_cred_t authz_cred_flag,
    lcas_request_t request,
    lcas_cred_id_t lcas_cred,
    X509 * px509_cred_in,
    STACK_OF(X509) * px509_chain_in
)
{
    char * logstr = NULL;
    char *              dn              = NULL; 
    struct vomsdata *   vd              = NULL;
    int                 errNo           = 0;
    gss_cred_id_t       cred            = GSS_C_NO_CREDENTIAL;
    X509 *              px509_cred      = NULL;
    STACK_OF(X509) *    px509_chain     = NULL;
    X509 *              px509_cred_gss  = NULL;
    STACK_OF(X509) *    px509_chain_gss = NULL;
    GACLuser *          gacluser        = NULL;
    int                 rc              = 0;
    char *              dummy           = NULL;

#ifdef TESTBIO
    int err;
    int res;
    BIO *in = NULL;
    X509 *x = NULL;
    STACK_OF(X509) *chain;
#endif

    if (authz_cred_flag == USE_X509)
        logstr = "\tlcas_plugin_voms-plugin_confirm_authorization_from_x509()";
    else if (authz_cred_flag == USE_GSS)
        logstr = "\tlcas_plugin_voms-plugin_confirm_authorization()";
    else
    {
        logstr = "\tlcas_plugin_voms-_plugin_confirm_authorization()";
        lcas_log(0, "%s: Unknown authZ credential flag: %d (failure)\n",
                 logstr, authz_cred_flag);
        goto fail_voms;
    }

    /*
     * The beginning
     */
    lcas_log_debug(1,"%s:\n", logstr);

    /*
     * Get the gss_cred_t credential and get the globus name
     */
    if ( (dn = lcas_get_dn(lcas_cred)) == NULL)
    {
        lcas_log(0, "%s: Error, user DN empty\n", logstr);
        goto fail_voms;
    }
    if (authz_cred_flag == USE_X509)
    {
        if (!px509_cred_in || !px509_chain_in)
        {
            lcas_log(0, "%s: empty X509 and or chain presented (failure)\n", logstr);
            goto fail_voms;
        }
        px509_cred = px509_cred_in;
        px509_chain = px509_chain_in;
    }
    else if (authz_cred_flag == USE_GSS)
    {
        /* Get the GSS credential from the lcas_cred */
        if ( (cred = lcas_get_gss_cred(lcas_cred)) == GSS_C_NO_CREDENTIAL)
        {
            lcas_log(0, "%s: Error, user gss credential is empty !\n", logstr);
            goto fail_voms;
        }
        else
        {
            /* Abondoned feature to throw the credentials to a file */
            #undef EXPORT_CREDENTIAL
            #if EXPORT_CREDENTIAL
            if (exportCredentials (cred))
                goto fail_voms;
            #endif /* EXPORT_CREDENTIAL */

            /*
             * in USE_GSS mode:
             * Retrieve a newly created X509 struct and X509 chain from gss credential (should be freed)
             */
            if (extractX509fromGSS (cred, px509_cred_gss, px509_chain_gss))
            {
                /* Extraction failed, appropriate error message is in the function */
                goto fail_voms;
            }
        }
    }



    lcas_log_debug(1,"%s: vomsdir = %s\n", logstr, vomsdir);
    lcas_log_debug(1,"%s: certdir = %s\n", logstr, certdir);
    if ((vd = VOMS_Init(vomsdir, certdir)) == NULL)
    {
        lcas_log(0,"%s: Failed to initialize voms data structure\n", logstr);
        lcas_log(0,"%s:  This may be because either the specified voms directory (%s)\n",logstr,vomsdir);
        lcas_log(0,"%s:  or the specified CA certificates directory (%s) does not exist\n", logstr, certdir);
        goto fail_voms;
    }
    lcas_log_debug(1,"%s: voms data structure initialized\n", logstr);

#ifdef TESTBIO
    in = BIO_new(BIO_s_file());
    chain = load_chain("/home/gridtest/cvs/fabric_mgt/gridification/lcas/modules/voms/x509up_u500");
    if (in)
    {
        if (BIO_read_filename(in, "/home/gridtest/cvs/fabric_mgt/gridification/lcas/modules/voms/x509up_u500") > 0)
        {
            x = PEM_read_bio_X509(in, NULL, 0, NULL);

            res = VOMS_Retrieve(x, chain, RECURSE_CHAIN, vd, &err);

            if (res)
                print_vomsdata(vd);
            else
                printf("%s: ERROR!\n", logstr);
        }
    }

    if (!res)
    {
        printf("%s: err: %s\n", logstr, retmsg[err]);
    }
#else

    if ( (debug_certs == 1) && (debug_printfile != NULL) )
    {
        char * output_single = NULL;
        char * output_chain  = NULL;

        output_single = malloc (sizeof (char) * (strlen(debug_printfile) + strlen(".single") + 1));
        output_chain  = malloc (sizeof (char) * (strlen(debug_printfile) + strlen(".chain") + 1));

        strcpy (output_single, debug_printfile);
        strcpy (output_chain , debug_printfile);

        strcat(output_single, ".single");
        strcat(output_chain,  ".chain");


        lcas_log(0,"DEBUG DEBUG, printing all the certificates in the chain to files: %s.{single|chain}\n", debug_printfile);

        lcas_print_x509_to_string (px509_cred, output_single);

        if (px509_chain)
        {
            STACK_OF(X509) *  dupChain     = NULL;
            X509 *            mylittlecert = NULL;

            dupChain = (STACK_OF(type) *) sk_X509_dup(px509_chain);

            while (mylittlecert = sk_X509_pop(dupChain))
            {
               lcas_print_x509_to_string (mylittlecert, output_chain);
            }

            if (dupChain) sk_X509_free(dupChain);
        }
        else
        {
            lcas_log(0,"%s: no input X509 chain!\n", logstr);
        }

        free(output_single);
        free(output_chain);

        lcas_log(0,"DEBUG DEBUG, done and continueing normal operation.\n");
    }


    if (VOMS_Retrieve(px509_cred, px509_chain, RECURSE_CHAIN, 
                         vd, &errNo))
#endif
    {
        lcas_vo_data_t *   lcas_vo_data = NULL;
        struct voms **     volist = vd->data;
        struct voms *      vo;
        char *             bufptr = NULL;
        int                k = 0;
        int                j = 0;
        int                foundmatch = 0;

        lcas_log_debug(1,"%s: We got something, errNo = %d\n", logstr, errNo);
        print_vomsdata(vd);

        if (authformat == GACL_FORMAT)
        {
            /* initialize gacl */
            GACLinit();
            if (lcas_gacl_add_dn(&gacluser, dn) != 0)
            {
                lcas_log(0,"%s: error adding dn %s to gacluser\n", logstr, dn);
                goto fail_voms;
            }
            else
            {
                lcas_log_debug(2,"%s: added dn %s to gacluser\n", logstr, dn);
            }
        }

        while(volist[k])
        {
            vo = volist[k++];
            lcas_log_debug(1,"%s: setting voms data for VO == %s\n", logstr,
                             vo->voname);
            lcas_log_debug(2,"%s: setting voms data for VO server == %s\n", logstr,
                             vo->server);

            switch (vo->type) {
                case TYPE_NODATA:
                    lcas_log_debug(1,"%s: NO DATA\n", logstr);
                    break;
                case TYPE_CUSTOM:
                    lcas_log_debug(1,"%s: %*s\n", logstr, vo->datalen - 10, vo->custom);
                    break;
                case TYPE_STD:
                    j = 0;
                    while (vo->std[j]) {
                        char *  fqan = NULL;
                        int     ifqan = 0;

                        lcas_vo_data=lcas_createVoData(vo->voname,vo->std[j]->group,
                                                           NULL, vo->std[j]->role, vo->std[j]->cap
                        );
                        if (! lcas_vo_data)
                        {
                            lcas_log(0,"%s: could not create VoData structure (failure)\n", logstr);
                            goto fail_voms;
                        }
                        if ( lcas_stringVoData(lcas_vo_data, voms_buffer, VOMS_BUFFER_SIZE) )
                        {
                            lcas_log(0,"%s: error in casting VoData structure into string (failure)\n", logstr);
                            goto fail_voms;
                        }
                        /*
                         * Now that we have the VO-GROUP-ROLE combination as a string it has to be matched in the
                         * authorization file. Check the format.
                         */
                        /*
                         * From version 1.2.x we support also FQANs. These are available in the as a NULL terminated
                         * list of strings: vo->fqan
                         * We're going to increment the counter of this array along with the counter of the triplet
                         * and check if the fqan is found when using the SIMPLE_FORMAT
                         */
                        if (vo->fqan)
                        {
                            while ((vo->fqan)[ifqan] != NULL)
                            {
                                if (ifqan == j) /* found the corresponding fqan */
                                {
                                    fqan = (vo->fqan)[ifqan];
                                    lcas_log_debug(1, "%s: fqan:   %s\n", logstr, fqan);
                                    break;
                                }
                                ifqan++;
                            }
                        }
                        if (fqan == NULL)
                        {
                            lcas_log(0, "%s: WARNING could not find fqan inside VOMS structure, only VOMS triplet\n", logstr);
                        }
                        if (authformat == SIMPLE_FORMAT)
                        {
                            /* check with gridlist, first try it with the fqan */
                            rc = lcas_gridlist(fqan, &dummy, authfile, MATCH_ONLY_DN|MATCH_WILD_CHARS, NULL, NULL);
                            if ( rc == LCAS_MOD_ENTRY )
                            {
                                /* Entry found for VO combi */
                                lcas_log_debug(1,"%s: entry found for fqan %s\n", logstr, fqan);
                                foundmatch++;
                            }
                            else if ( rc == LCAS_MOD_NOFILE )
                            {
                                /* file not found */
                                lcas_log(0, "%s: Error, Cannot find authorization file: %s\n", logstr, authfile);
                                if (dummy)
                                {
                                    free(dummy);
                                    dummy = NULL;
                                }
                                goto fail_voms;;
                            }
                            else
                            {
                                /* no entry found, not fatal */
                                lcas_log_debug(1, "%s: no entry for fqan %s in %s\n", logstr, fqan, authfile);
                            }
                            if (dummy)
                            {
                                free(dummy);
                                dummy = NULL;
                            }

                            /* If no entry found for the FQAN try it with the VO string (in voms_buffer) */
                            if (foundmatch == 0)
                            {
                                lcas_log_debug(1, "%s: Trying old format: %s instead of %s\n", logstr, voms_buffer, fqan);
                                rc = lcas_gridlist(voms_buffer, &dummy, authfile, MATCH_ONLY_DN|MATCH_WILD_CHARS, NULL, NULL);
                                if ( rc == LCAS_MOD_ENTRY )
                                {
                                    /* Entry found for VO combi */
                                    lcas_log_debug(1,"%s: entry found for VOMS triplet %s\n", logstr, voms_buffer);
                                    foundmatch++;
                                }
                                else if ( rc == LCAS_MOD_NOFILE )
                                {
                                    /* file not found */
                                    lcas_log(0, "%s: Error, Cannot find authorization file: %s\n", logstr, authfile);
                                    if (dummy)
                                    {
                                        free(dummy);
                                        dummy = NULL;
                                    }
                                    goto fail_voms;;
                                }
                                else
                                {
                                    /* no entry found, not fatal */
                                    lcas_log_debug(1, "%s: no entry for VOMS triplet %s in %s\n", logstr, voms_buffer, authfile);
                                }
                                if (dummy)
                                {
                                    free(dummy);
                                    dummy = NULL;
                                }
                            }
                        }
                        else if (authformat == GACL_FORMAT)
                        {
                            /* Fill the gacluser with another VOMS credential, do the ACL checking afterwards */
                            if ( (gacl_use_voms_dn == USE_VOMS_DN) || (gacl_use_voms_dn == ALWAYS_USE_VOMS_DN) )
                            {
                                if (!lcas_parseVostring(vo->server))
                                {
                                    lcas_log(0,"%s: No VOMS server DN found in user cert. (failure)\n", logstr);
                                    goto fail_voms;
                                }
                                /* First create a credential with the VOMS DN */
                                if (lcas_gacl_add_vomsdata(&gacluser, lcas_vo_data, vo->server) != 0)
                                {
                                    lcas_log(0,"%s: error adding voms data for %s to gacluser\n", logstr, voms_buffer);
                                    goto fail_voms;
                                }
                                else
                                {
                                    lcas_log_debug(2,"%s: added voms data for %s to gacluser\n", logstr, voms_buffer);
                                }
                                /* Then create a credential without the VOMS DN if (gacl_use_voms_dn == USE_VOMS_DN) */
                                if (gacl_use_voms_dn == USE_VOMS_DN)
                                {
                                    if (lcas_gacl_add_vomsdata(&gacluser, lcas_vo_data, NULL) != 0)
                                    {
                                        lcas_log(0,"%s: error adding voms data for %s to gacluser\n", logstr, voms_buffer);
                                        goto fail_voms;
                                    }
                                    else
                                    {
                                        lcas_log_debug(2,"%s: added voms data for %s to gacluser\n", logstr, voms_buffer);
                                    }
                                }
                            }
                            else if (gacl_use_voms_dn == DONT_USE_VOMS_DN)
                            {
                                /* Create a credential without the VOMS DN */
                                if (lcas_gacl_add_vomsdata(&gacluser, lcas_vo_data, NULL) != 0)
                                {
                                    lcas_log(0,"%s: error adding voms data for %s to gacluser\n", logstr, voms_buffer);
                                    goto fail_voms;
                                }
                                else
                                {
                                    lcas_log_debug(2,"%s: added voms data for %s to gacluser\n", logstr, voms_buffer);
                                }
                            }
                        }
                        else if (authformat == XACML_FORMAT)
                        {
                            /* check with xacml */
                            lcas_log(0,"%s: This (%d) format is not supported yet\n", logstr, authformat);
                        }
                        else
                        {
                            lcas_log(0,"%s: This (%d) format is not supported\n", logstr, authformat);
                        }
                        /* Add credential */
                        /* copy address of voms_buffer[0] in bufptr, because you cannot take the address of the array voms_buffer */
                        bufptr = voms_buffer;
//                        addCredentialData(LCAS_VO_CRED_STRING, (void *) &bufptr);
//                        addCredentialData(LCAS_VO_CRED, (void *) lcas_vo_data);
                        if ( lcas_deleteVoData(&lcas_vo_data) )
                        {
                            lcas_log(0,"%s: error while deleting VoData structure (failure)\n", logstr);
                            goto fail_voms;
                        }
                        j++;
                    }
                    break;
            }
        }
        lcas_log_debug(1,"%s: doing VOMS_Destroy\n", logstr);
        VOMS_Destroy(vd);
        lcas_log_debug(1,"%s: done\n", logstr);

        /* evaluate the matching, may become more elaborate */
        
        if (authformat == GACL_FORMAT)
        {
            /* check with gacl */
            /* Perhaps do something with fatalness, if deny found always quit */
            rc = lcas_check_gacl(gacluser, authfile);
            if (rc == 0)
            {
                lcas_log_debug(1, "%s: authorization granted based on VOMS info for user %s in %s\n", logstr, dn, authfile);
            }
            else
            {
                /* no authorization --> failure */
                lcas_log(0, "%s: authorization denied based on VOMS info for user %s in %s\n", logstr, dn, authfile);
                goto fail_voms;
            }
        }
        else
        {
            if (foundmatch)
            {
                lcas_log_debug(1,"%s: found a matching VO entry in the authorization file\n", logstr);
            }
            else
            {
                lcas_log(0,"%s: Did not find a matching VO entry in the authorization file\n", logstr);
                goto fail_voms;
            }
        }
        
    }
#ifdef TESTBIO
#else
    else if (errNo == VERR_NOEXT)
    {
        /*
         * If no voms info is found in user proxy: try with the user dn if
         * use_user_dn flag is specified
         */
        if (use_user_dn != 0)
        {
            lcas_log_debug(1,"%s: VOMS extensions missing from certificate, trying\n", logstr);
            lcas_log_debug(1,"%s: to find user dn (%s) in authorization file %s\n", logstr, dn, authfile);

            if (authformat == SIMPLE_FORMAT)
            {
                /* check with gridlist */
                rc = lcas_gridlist(dn, &dummy, authfile, MATCH_ONLY_DN, NULL, NULL);
                if ( rc == LCAS_MOD_ENTRY )
                {
                    /* Entry found for user DN */
                    lcas_log_debug(1,"%s: entry found for %s\n", logstr, dn);
                }
                else if ( rc == LCAS_MOD_NOFILE )
                {
                    /* file not found */
                    lcas_log(0, "%s: Error, Cannot find authorization file: %s\n", logstr, authfile);
                    if (dummy)
                    {
                        free(dummy);
                        dummy = NULL;
                    }
                    goto fail_voms;;
                }
                else
                {
                    /* no entry found --> failure */
                    lcas_log(0, "%s: authorization denied based on DN info for user\n", logstr);
                    lcas_log(0, "%s: %s in %s\n", logstr, dn, authfile);
                    lcas_log(0, "%s: (in addition no VOMS info was found in user proxy)\n", logstr);
                    if (dummy)
                    {
                        free(dummy);
                        dummy = NULL;
                    }
                    goto fail_voms;
                }
            }
            else if (authformat == GACL_FORMAT)
            {
                /* initialize gacl */
                GACLinit();
                if (lcas_gacl_add_dn(&gacluser, dn) != 0)
                {
                    lcas_log(0,"%s: error adding dn %s to gacluser\n", logstr, dn);
                    goto fail_voms;
                }
                else
                {
                    lcas_log_debug(2,"%s: added dn %s to gacluser\n", logstr, dn);
                }
                /* check with gacl */
                /* Perhaps do something with fatalness, if deny found always quit */
                rc = lcas_check_gacl(gacluser, authfile);
                if (rc == 0)
                {
                    lcas_log_debug(1, "%s: authorization granted based on DN info for user %s in %s\n", logstr, dn, authfile);
                }
                else
                {
                    /* no authorization --> failure */
                    lcas_log(0, "%s: authorization denied based on DN info for user\n", logstr);
                    lcas_log(0, "%s: %s in %s\n", logstr, dn, authfile);
                    lcas_log(0, "%s: (in addition no VOMS info was found in user proxy)\n", logstr);
                    goto fail_voms;
                }
            }
            else
            {
                lcas_log(0,"%s: This (%d) format is not supported yet, but this should have been checked earlier in the process! \n", logstr, authformat);
            }
        }
        else
        {
            lcas_log(0,"%s: VOMS extensions missing from certificate (failure)!\n", logstr);
            goto fail_voms;
        }
    }
    else if (errNo == VERR_IDCHECK)
    {
        lcas_log(0,"%s: VOMS User data in extension different from the real ones (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_TIME)
    {
        lcas_log(0,"%s: VOMS extensions expired for at least one of the VOs (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_ORDER)
    {
        lcas_log(0,"%s: The ordering of the VOMS groups, as required by the client, was not delivered by VOMS (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_NOSOCKET)
    {
        lcas_log(0,"%s: VOMS Socket problem (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_NOIDENT)
    {
        lcas_log(0,"%s: VOMS Cannot identify itself (certificate problem) (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_COMM)
    {
        lcas_log(0,"%s: VOMS server problem (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_PARAM)
    {
        lcas_log(0,"%s: Wrong parameters for VOMS (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_NOINIT)
    {
        lcas_log(0,"%s: VOMS initialization error (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_EXTRAINFO)
    {
        lcas_log(0,"%s: VO name and URI missing (in proxy ?) (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_FORMAT)
    {
        lcas_log(0,"%s: Wrong VOMS data format (in proxy ?) (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_NODATA)
    {
        lcas_log(0,"%s: Empty VOMS extension (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_PARSE)
    {
        lcas_log(0,"%s: VOMS parse error (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_DIR)
    {
        lcas_log(0,"%s: VOMS directory error (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_SIGN)
    {
        lcas_log(0,"%s: VOMS Signature error (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_SERVER)
    {
        lcas_log(0,"%s: Unidentifiable VOMS server (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_MEM)
    {
        lcas_log(0,"%s: Memory problems in VOMS_Retrieve() (failure)!\n", logstr);
        goto fail_voms;
    }
    else if (errNo == VERR_VERIFY)
    {
        char * err_buffer = VOMS_ErrorMessage (vd, VERR_VERIFY, NULL, 0);
        lcas_log(0,"%s: Generic verification error for VOMS (failure): %s\n", logstr, err_buffer);
        free (err_buffer);
        goto fail_voms;
    }
    else if (errNo == VERR_TYPE)
    {
        lcas_log(0,"%s: Returned VOMS data of unknown type (failure)!\n", logstr);
        goto fail_voms;
    }
    else
    {
        lcas_log(0,"%s: VOMS_Retrieve() error --> %d (failure)!\n", logstr, errNo);
        goto fail_voms;
    }
#endif

    /* succes */
// success_voms:
    if (px509_cred_gss) X509_free(px509_cred_gss);
    if (px509_chain_gss) sk_X509_pop_free(px509_chain_gss, X509_free);
    if (authformat == GACL_FORMAT)
    {
        if (gacluser)
        {
            GACLfreeUser(gacluser);
            gacluser = NULL;
        }
    }
    lcas_log_time(0,"%s: voms plugin succeeded\n", logstr);
    return LCAS_MOD_SUCCESS;

 fail_voms:
    if (px509_cred_gss) X509_free(px509_cred_gss);
    if (px509_chain_gss) sk_X509_pop_free(px509_chain_gss, X509_free);
    if (authformat == GACL_FORMAT)
    {
        if (gacluser)
        {
            GACLfreeUser(gacluser);
            gacluser = NULL;
        }
    }
    lcas_log_time(0,"%s: voms plugin failed\n", logstr);
    return LCAS_MOD_FAIL;
}

/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    char * logstr = "\tlcas_plugin_voms-plugin_terminate()";
    lcas_log_debug(1,"%s: terminating\n", logstr);
    if (vomsdir)
    {
        free(vomsdir);
        vomsdir = NULL;
    }
    if (certdir)
    {
        free(certdir);
        certdir = NULL;
    }
    if (authfile)
    {
        free(authfile);
        authfile = NULL;
    }

    return LCAS_MOD_SUCCESS;
}


/******************************************************************************
Function:   lcas_gacl_add_dn
Description:
    Add the user DN to the gacl user credential
Parameters:
    pgacluser:      pointer to the gacl user
    user_dn:        the DN of the user

Returns:
    0 : succes
    1 : failure
******************************************************************************/
/*!
    \fn lcas_gacl_add_dn(
        GACLuser ** pgacluser,
        char *      user_dn
        )
    \brief Add the user_dn to gacluser
    
    Add the user_dn to gacluser. If the gacluser does not exist, it will be created.

    \param pgacluser      pointer to the gacl user
    \param user_dn        the DN of the user
    \retval 0             success
    \retval 1             failure
*/
static int lcas_gacl_add_dn(
        GACLuser ** pgacluser,
        char *      user_dn
)
{
    char *      logstr = "\tlcas_plugin_voms-lcas_gacl_add_dn()";
    GACLcred *  dn_cred = NULL;
    GACLuser *  gacluser = NULL;
    
    if (user_dn)
    {
        if ( (dn_cred = GACLnewCred("person")) != NULL )
        {
            lcas_log_debug(3,"%s: adding dn = %s to dn_cred\n", logstr, user_dn);
            GACLaddToCred(dn_cred, "dn", user_dn);
        }
        else
        {
            lcas_log_debug(1,"%s: Cannot create new credential\n", logstr);
            return 1;
        }
    }
    else
    {
        lcas_log_debug(1,"%s: empty user DN !, cannot fill GACLuser\n", logstr);
        return 1;
    }

    /* if necessary create user and add credential to user */
    if (pgacluser == NULL)
    {
        /* This should not happen ! */
        lcas_log(0, "%s: ptr to gacluser is NULL !, wrong invocation of lcas_gacl_add_dn()\n", logstr);
        GACLfreeCred(dn_cred);
        return 1;
    }
    gacluser = *pgacluser;
    if (gacluser == NULL) /* create new user */
    {
        if ( (gacluser = GACLnewUser(dn_cred)) == NULL )
        {
            lcas_log(0, "%s: Could not create new user\n", logstr);
            GACLfreeCred(dn_cred);
            return 1;
        }
        *pgacluser = gacluser;
    }
    else
    {
        if (GACLuserAddCred(gacluser, dn_cred) != 1)
        {
            lcas_log(0, "%s: Could not add credential to user\n", logstr);
            return 1;
        }
    } 

    /* clean credential */
    if (dn_cred)
    {
        /* Cannot free credential, because it's used in the gacluser
           structure (only a copy of the credential pointer is made)
        GACLfreeCred(dn_cred);
        */
        dn_cred = NULL;
    }
    return 0;
}


/******************************************************************************
Function:   lcas_gacl_add_vomsdata
Description:
    Add the VOMS data to the gacl user
Parameters:
    pgacluser:       pointer to the gacl user
    lcas_voms_data:  the gathered VOMS data structure
    voms_server_dn:  the DN of the VOMS server that signed the VOMS certificate

Returns:
    0 : succes
    1 : failure
******************************************************************************/
/*!
    \fn lcas_gacl_add_vomsdata(
        GACLuser **       pgacluser,
        lcas_vo_data_t *  lcas_voms_data,
        char *            voms_server_dn
        )
    \brief Add the VOMS data to the gacl user
    
    Add the VOMS data to the gacl user. If the gacluser does not exist, it will be created.

    \param pgacluser       pointer to the gacl user
    \param lcas_voms_data  the gathered VOMS data structure
    \param voms_server_dn  the DN of the VOMS server that signed the VOMS certificate

    \retval 0              success
    \retval 1              failure
*/
static int lcas_gacl_add_vomsdata(
        GACLuser **       pgacluser,
        lcas_vo_data_t *  lcas_voms_data,
        char *            voms_server_dn
)
{
    char *      logstr = "\tlcas_plugin_voms-lcas_gacl_add_vomsdata()";
    GACLcred *  voms_cred = NULL;
    GACLuser *  gacluser = NULL;
    
    if (lcas_voms_data)
    {
        if ( (voms_cred = GACLnewCred("voms-cred")) != NULL )
        {
            char * strptr = NULL;

            if ( (strptr = lcas_parseVostring(voms_server_dn)) )
            {
                lcas_log_debug(3,"%s: adding voms       = %s to voms_cred\n", logstr, strptr);
                GACLaddToCred(voms_cred, "voms", strptr);
            }
            if ( (strptr = lcas_parseVostring(lcas_voms_data->vo)) )
            {
                lcas_log_debug(3,"%s: adding vo         = %s to voms_cred\n", logstr, strptr);
                GACLaddToCred(voms_cred, "vo", strptr);
            }
            if ( (strptr = lcas_parseVostring(lcas_voms_data->group)) )
            {
                lcas_log_debug(3,"%s: adding group      = %s to voms_cred\n", logstr, strptr);
                GACLaddToCred(voms_cred, "group", strptr);
            }
            if ( (strptr = lcas_parseVostring(lcas_voms_data->subgroup)) )
            {
                lcas_log_debug(3,"%s: adding subgroup   = %s to voms_cred\n", logstr, strptr);
                GACLaddToCred(voms_cred, "group", strptr);
            }
            if ( (strptr = lcas_parseVostring(lcas_voms_data->role)) )
            {
                lcas_log_debug(3,"%s: adding role       = %s to voms_cred\n", logstr, strptr);
                GACLaddToCred(voms_cred, "role", strptr);
            }
            if ( (strptr = lcas_parseVostring(lcas_voms_data->capability)) )
            {
                lcas_log_debug(3,"%s: adding capability = %s to voms_cred\n", logstr, strptr);
                GACLaddToCred(voms_cred, "capability", strptr);
            }
        }
        else
        {
            lcas_log_debug(1,"%s: Cannot create new credential\n", logstr);
            return 1;
        }
    }
    else
    {
        lcas_log_debug(1,"%s: empty VO data !, cannot fill GACLuser\n", logstr);
        return 1;
    }

    /* if necessary create user and add credential to user */
    if (pgacluser == NULL)
    {
        /* This should not happen ! */
        lcas_log(0, "%s: ptr to gacluser is NULL !, wrong invocation of lcas_gacl_add_vomsdata()\n", logstr);
        GACLfreeCred(voms_cred);
        return 1;
    }
    gacluser = *pgacluser;
    if (gacluser == NULL) /* create new user */
    {
        if ( (gacluser = GACLnewUser(voms_cred)) == NULL )
        {
            lcas_log(0, "%s: Could not create new user\n", logstr);
            GACLfreeCred(voms_cred);
            return 1;
        }
        *pgacluser = gacluser;
    }
    else
    {
        if (GACLuserAddCred(gacluser, voms_cred) != 1)
        {
            lcas_log(0, "%s: Could not add credential to user\n", logstr);
            return 1;
        }
    } 

    /* clean credential */
    if (voms_cred)
    {
        /* Cannot free credential, because it's used in the gacluser
           structure (only a copy of the credential pointer is made)
        GACLfreeCred(voms_cred);
        */
        voms_cred = NULL;
    }
    return 0;
}


/******************************************************************************
Function:   lcas_check_gacl
Description:
    Apply the LCAS authorization GACL to the GACL user
Parameters:
    gacluser:  the gacluser, which consists of his DN and VOMS entries
    gaclfile:  the file containing the LCAS GACL.

Returns:
    0 : succes
    1 : failure
******************************************************************************/
/*!
    \fn lcas_check_gacl(
        GACLuser *  gacluser,
        char *      gaclfile
        )
    \brief check the LCAS GACL

    Apply the LCAS authorization GACL to the user credentials (VOMS cred. and DN)

    \param gacluser  the gacl_user, which consists of his DN and VOMS entries
    \param gaclfile  the file containing the LCAS GACL
    \retval 0        success
    \retval 1        failure
*/
static int lcas_check_gacl(
        GACLuser *  gacluser,
        char *      gaclfile
)
{
    /*
     * Procedure:
     * 1. initialize gacl
     * 2. load the gacl file
     * 3. create an GACLuser (list of credentials)
     * 4. check the premissions for the user
     *    (for the moment if the result is none --> deny, otherwise allow)
     */

    char *      logstr = "\tlcas_plugin_voms-lcas_check_gacl()";
    GACLacl *   lcasacl = NULL;
    GACLperm    lcas_gacl_perm = GACL_PERM_NONE;

    /* load the gacl file */
    lcasacl = GACLloadAcl(gaclfile);
    lcas_log_debug(1,"%s: gacl authorization file %s loaded\n", logstr, gaclfile);
    if (lcasacl != NULL)
    {
//        GACLprintAcl(lcasacl, stderr);
    }
    else
    {
        lcas_log(0,"%s: lcasacl is NULL\n", logstr);
        return 1;
    }

    /* check the premissions for the user */
    lcas_gacl_perm = GACLtestExclAcl(lcasacl, gacluser);
    lcas_log_debug(1,"%s: exclusive permission = %d\n", logstr, lcas_gacl_perm);

    lcas_gacl_perm = GACLtestUserAcl(lcasacl, gacluser);
    lcas_log_debug(1,"%s: permission = %d\n", logstr, lcas_gacl_perm);
    if (GACLhasNone(lcas_gacl_perm))
    {
        if (lcas_get_debug_level() >= 5)
            GACLprintAcl(lcasacl, stderr);
        GACLfreeAcl(lcasacl);
        return 1;
    }
    else
    {
        GACLfreeAcl(lcasacl);
        return 0;
    }
}

static void print_vomsdata(struct vomsdata *d)
{
    char * logstr = "\tlcas_plugin_voms-print_vomsdata()";
    struct voms **vo = d->data;
    struct voms *v;
    int k = 0;
    int j =0;
  
    while(vo[k])
    {
        v = vo[k++];
        lcas_log_debug(1,"%s: %d *******************************************\n", logstr,k);
        lcas_log_debug(1,"%s: SIGLEN: %d\n", logstr, v->siglen);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): USER:   %s\n", v->user);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): UCA:    %s\n", v->userca);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): SERVER: %s\n", v->server);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): SCA:    %s\n", v->serverca);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): VO:     %s\n", v->voname);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): URI:    %s\n", v->uri);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): DATE1:  %s\n", v->date1);
        lcas_log_a_string_debug(1, "\tlcas_plugin_voms-print_vomsdata(): DATE2:  %s\n", v->date2);

        switch (v->type)
        {
        case TYPE_NODATA:
            lcas_log_debug(1,"%s: NO DATA\n", logstr);
            break;
        case TYPE_CUSTOM:
            lcas_log_debug(1,"%s: VOMS custom type. Wont print.\n", logstr);
            break;
        case TYPE_STD:
            j = 0;
            if (v->fqan)
            {
                while ((v->fqan)[j] != NULL)
                {
                    lcas_log_a_string_debug(1,
                        "\tlcas_plugin_voms-print_vomsdata(): fqan:   %s\n",
                        (v->fqan)[j]);
                    j++;
                }
            }
            j = 0;
            if (v->std)
            {
                while (v->std[j])
                {
                    lcas_log_a_string_debug(1,
                        "\tlcas_plugin_voms-print_vomsdata(): GROUP:  %s\n", v->std[j]->group);
                    lcas_log_a_string_debug(1,
                        "\tlcas_plugin_voms-print_vomsdata(): ROLE:   %s\n", v->std[j]->role);
                    lcas_log_a_string_debug(1,
                        "\tlcas_plugin_voms-print_vomsdata(): CAP:    %s\n", v->std[j]->cap);
                    j++;
                }
            }
            break;
        }
        lcas_log_debug(1,"%s: %d *******************************************\n", logstr,k);
    }

    if (d->workvo)
    {
        lcas_log_a_string_debug(1,
            "\tlcas_plugin_voms-print_vomsdata(): WORKVO: %s\n", d->workvo);
    }

    if (d->extra_data)
    {
        lcas_log_a_string_debug(1,
            "\tlcas_plugin_voms-print_vomsdata(): EXTRA: %s\n", d->extra_data);
    }
}


/******************************************************************************
Function:    lcas_print_x509_to_string
Description:
    Extract from X509 certificate and
        write the output to a file.
    Simulair to `openssl x509 -in <file.pem> -text -noout

    Note: only for debugging purposes
Parameters:
    px509:    pointer to a x509 cert
Returns:
    void
******************************************************************************/
void lcas_print_x509_to_string (X509 * px509, char * output_file)
{
    char *            logstr = "\tlcas_print_x509_to_string()";

    if (px509)
    {
        FILE *            tmpFile = NULL;

        lcas_log(0,"%s\n", logstr);

        tmpFile = fopen(output_file, "a");
        X509_print_fp(tmpFile, px509);
        fclose (tmpFile);
    }
    else
    {
        lcas_log(0,"%s: no input X509 cert!\n", logstr);
    }
}



/******************************************************************************
CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/lcas-plugins-voms/src/voms/lcas_voms.c,v $
    $Date: 2010-05-03 10:44:38 $
    $Revision: 1.15 $
    $Author: okoeroo $
******************************************************************************/
